#!/usr/bin/python3
import re
import sys
from decimal import Decimal
from pprint import pprint

""" cleanup_gcode.py <power> <feed> <gcode> 

    cleans up gcode generated by Inkscape's GCodeTools
    for use on a laser-cutter without a z-axis
    
    Copyright (C) 2024 Andy D'Arcy Jewell and MakerPunkBuzz
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

feed = sys.argv[1]
power = sys.argv[2]
gcode = open(sys.argv[3]).readlines()
outfile = 'fixed.'+sys.argv[3]
objfile = 'objects.'+sys.argv[3]
MAX_X = 100000
MAX_Y = 100000

remove_lines = ( '', 'G00 Z0.000000', 'G01 Z0.000000 F100.0(Penetrate)', )
replace_strings = ( (r' Z[-.0-9]+',''), (r' F[.0-9]+',f' F{feed} S{power}'), )
add_before = { '(Header end.)':'M08 (Air on)', 
            }

def remove_comments(text):
    text = re.sub(r';.*$','',text)
    text = re.sub(r'\([^)]+\)','',text)
    return text
    
def decode_block(text):
    text = remove_comments(text)
    if not text.strip(): return {}
#    print(text.split(' '))
    return dict( (code[0],Decimal(code[1:])) for code in text.split(' ')
                if code.strip() != '' )

def enumerate_objects(gcode):    
    # fix cut order so that small cuts inside larger cuts are done first
    objects = []
    header = []
    footer = []
    # find objects
    mode = "header"
    for i,line in enumerate(gcode):
#        print(mode,repr(line))
        if mode == "header":
            if gcode[i+1].startswith('(Start cutting path id:'):
                mode = "find"
            header.append(line)
        elif mode == "find":
            if line.startswith('(Start cutting path id:'):
                mode = "path"
                pathid = line[line.find(':')+1:line.rfind(')')]
                print(line,pathid)
                obj = dict(
                    pathid = pathid,
                    lines = [line,],
                    points = [],
                    minx = Decimal(MAX_X), # Init to obviously wrong values...
                    maxx = Decimal(0),
                    miny = Decimal(MAX_Y),
                    maxy = Decimal(0),
                    )
                objects.append(obj)
            elif line.startswith('(Footer)'):
                mode = "footer"                
                footer.append(line)
        elif mode == "path":
            obj['lines'].append(line)
            if line.startswith('(End cutting path id:'):
                mode = "find"
                print(f'(End cutting path id:{pathid})')
            else:
                block = decode_block(line)
                if ('G' in block and 
                    block['G'] in (Decimal('00'),Decimal('01'),
                                Decimal('02'),Decimal('03'))
                                ):
                    obj['points'].append((block['X'],block['Y']))
                    if block['X'] > obj['maxx']: 
                        obj['maxx'] = block['X']
                    elif block['X'] < obj['minx']: 
                        obj['minx'] = block['X']
                    if block['Y'] > obj['maxy']: 
                        obj['maxy'] = block['Y']
                    elif block['Y'] < obj['miny']: 
                        obj['miny'] = block['Y']
        elif mode == "footer":
            footer.append(line)

    # sort objects by bounding box, so that
    # objects contained within another object are cut first
    
    #shortcuts to save typing and visual noise
    maxx='maxx'
    minx='minx'
    maxy='maxy'
    miny='miny'
    pathid='pathid'
    print('objects')
    for o,obj in enumerate(objects):
        print(o,obj)
    vectors = list(reversed(sorted([ (o[minx],o[miny],o[maxx],o[maxy],o[pathid]) 
                    for o in objects ])))
    print('vectors')
    pprint(vectors)
    obj_map = dict( (o[pathid],o) for o in objects )
    print('obj_map')
    pprint(obj_map)
    new_objects = [ obj_map[o[4]] for o in vectors ]                    
#    print('new_objects')
#    pprint(new_objects)
        
    return new_objects,header,footer
    
fixed_gcode = []
for line in gcode:
    line_stripped = line.strip()
    if line_stripped not in remove_lines:
        if line_stripped in add_before:
            fixed_gcode.append(add_before[line_stripped]+'\n')
        for remove,replace in replace_strings:
            line = re.sub(remove,replace,line)
        fixed_gcode.append(line)

objects,header,footer = enumerate_objects(fixed_gcode)


pprint(objects)
reordered_gcode = header + [ l for o in objects for l in o['lines']] + footer

#print('reordered_gcode')
#pprint(reordered_gcode)



with open(objfile,"w") as objf:
    pprint(objects,objf)
    
with open(outfile,"w") as outf:
    outf.writelines(reordered_gcode)

